use crate::entities::{prelude::*, *};
use crate::state;
use crate::{error::AppError, views};
use argon2::password_hash::rand_core::OsRng;
use argon2::password_hash::SaltString;
use argon2::{Argon2, PasswordHasher};
use axum::extract::State;
use axum::{http::StatusCode, response::Html, routing::get, Form, Router};
use axum_csrf::{CsrfConfig, CsrfLayer, CsrfToken};
use base64::prelude::*;
use maud::html;
use notify::Watcher;
use sea_orm::*;
use serde::Deserialize;
use std::sync::Arc;
use std::{env, path::Path};
use tower_http::services::ServeDir;
use tower_livereload::LiveReloadLayer;
use tracing::{info, instrument};

#[instrument]
pub async fn new() -> Result<Router, anyhow::Error> {
    let app_router = Router::new()
        .route("/hello-world", get(views::greet_world))
        .route("/hello-world-text", get(views::greet_world_text));

    let assets_dir = ServeDir::new("./assets");
    let state = state::State::new().await?;

    let live_reload_layer = LiveReloadLayer::new();
    let reloader = live_reload_layer.reloader();

    let mut watcher = notify::recommended_watcher(move |_| reloader.reload())?;
    watcher.watch(Path::new("assets"), notify::RecursiveMode::Recursive)?;

    let cookie_key_bytes: Vec<u8> = BASE64_URL_SAFE_NO_PAD
        .decode(env::var("COOKIE_KEY").expect("COOKIE_KEY not set"))
        .expect("COOKIE_KEY not base64 URL-safe encoded");
    let cookie_key = cookie::Key::from(&cookie_key_bytes);
    let csrf_config = CsrfConfig::default().with_key(Some(cookie_key));

    let router = Router::new()
        .fallback(|| async { (StatusCode::NOT_FOUND, "404 page not found") })
        .nest("/app", app_router)
        .nest_service("/assets", assets_dir)
        .route("/", get(views::index))
        .route("/login", get(views::login).post(login))
        .route("/register", get(views::register).post(register))
        .route("/all_users", get(views::all_users))
        .with_state(state)
        .layer(CsrfLayer::new(csrf_config))
        .layer(live_reload_layer);

    Ok(router)
}

fn csrf_verify(c: CsrfToken, t: &str) -> Result<(), AppError> {
    // TODO: https://docs.rs/axum_csrf/latest/axum_csrf/#prevent-post-replay-attacks-with-csrf
    c.verify(t)?;
    Ok(())
}

type AppRes = Result<(StatusCode, Html<String>), AppError>;

#[derive(Deserialize)]
struct Register {
    authenticity_token: String,
    username: String,
    password: String,
}

impl TryInto<user::ActiveModel> for Register {
    type Error = AppError;

    fn try_into(self) -> Result<user::ActiveModel, Self::Error> {
        Ok(user::ActiveModel {
            id: ActiveValue::Set(uuid::Uuid::now_v7().into()),
            name: ActiveValue::Set(None),
            username: ActiveValue::Set(self.username),
            password_digest: ActiveValue::Set(password_digest(self.password)?),
        })
    }
}

fn password_digest<S>(s: S) -> Result<String, password_hash::Error>
where
    S: AsRef<str>,
{
    Ok(Argon2::default()
        .hash_password(s.as_ref().as_bytes(), &SaltString::generate(&mut OsRng))?
        .serialize()
        .to_string())
}

impl user::ActiveModel {}

async fn register(
    c: CsrfToken,
    State(s): State<Arc<state::State>>,
    Form(f): Form<Register>,
) -> AppRes {
    csrf_verify(c, &f.authenticity_token)?;

    // TODO: handle duplicate username

    let new: user::ActiveModel = f.try_into()?;
    let res = User::insert(new).exec(&s.db).await;

    info!("insert new user result: {:?}", res);

    res.map_err(anyhow::Error::from)?;

    Ok((
        StatusCode::CREATED,
        Html(
            html! {
                // h1 { (new_user.username) }
            }
            .into_string(),
        ),
    ))
}

#[derive(Deserialize)]
struct Login {
    authenticity_token: String,
    username: String,
    password: String,
}

async fn login(c: CsrfToken, Form(f): Form<Login>) -> AppRes {
    csrf_verify(c, &f.authenticity_token)?;

    Ok((
        StatusCode::OK,
        Html(
            html! {
                h1 { (f.username) }
                h1 { (f.password) }
            }
            .into_string(),
        ),
    ))
}
