use crate::entities::{prelude::*, *};
use crate::{
    error::AppError,
    partials::{footer, header},
    state,
};
use axum::{
    extract::State,
    http::StatusCode,
    response::{Html, IntoResponse},
};
use axum_csrf::CsrfToken;
use maud::html;
use sea_orm::EntityTrait;
use std::sync::Arc;
use tracing::instrument;

pub async fn csrf<F>(csrf: CsrfToken, cb: F) -> impl IntoResponse
where
    F: Fn(&str) -> Html<String>,
{
    let token = csrf.authenticity_token().unwrap();
    (csrf, cb(&token))
}

type AppRes = Result<(StatusCode, Html<String>), AppError>;

#[instrument]
pub async fn index() -> Html<String> {
    Html(
        html! {
            (header())
            main class="prose" {
                h1 { "Manage live lyrics and music displays" }
                p { "Stop struggling to share the same messy set of PowerPoint files or Google Presentations. Make editing and controlling your live lyrics and music displays easy and simple." }
                ul {
                    li { "Live collaboration with your team" }
                    li { "Fully compatible with any device" }
                    li { "Simple workflow" }
                    li { "Dark theme and light theme" }
                    li { "Generous free plan" }
                    li { "Lightweight and fast" }
                }
                section class="flex gap" {
                    a href="/register" class="button bg-primary" { "Try now" }
                    a class="button" href="/login" { "Login" }
                }
            }
            (footer())
        }
        .into_string(),
    )
}

pub async fn register(t: CsrfToken) -> impl IntoResponse {
    csrf(t, |token| {
        Html(
            html! {
                (header())
                main class="prose" {
                    h1 { "Register an account" }
                    form method="post" {
                        input type="hidden" name="authenticity_token" value=(token) {}
                        label {
                            "Username:"
                            input name="username" {}
                        }
                        label {
                            "Password:"
                            input type="password" name="password" {}
                        }
                        button type="submit" { "Create Account" }
                    }
                }
                (footer())
            }
            .into_string(),
        )
    })
    .await
}

pub async fn login(t: CsrfToken) -> impl IntoResponse {
    csrf(t, |token| {
        Html(
            html! {
                (header())
                main class="prose" {
                    h1 { "Login" }
                    form method="post" {
                        input type="hidden" name="authenticity_token" value=(token) {}
                        label {
                            input {}
                        }
                        label {
                            input {}
                        }
                    }
                }
                (footer())
            }
            .into_string(),
        )
    })
    .await
}

pub async fn all_users(State(s): State<Arc<state::State>>) -> Result<Html<String>, AppError> {
    let users: Vec<user::Model> = User::find().all(&s.db).await?;

    // @if let Some(name) = u.name {
    //     name
    // } @else {
    //     "N/A"
    // }
    Ok(Html(
        html! {
            (header())
            main class="prose" {
                h1 { "Users" }
                ul {
                    @if users.len() < 1 {
                        li { "It looks like there are no users yet!" }
                    } else {
                        @for u in users {
                            li {
                                (u.username)
                                @if let Some(name) = u.name {
                                    " ("
                                    (name)
                                    ")"
                                }
                            }
                        }
                    }
                }
            }
            (footer())
        }
        .into_string(),
    ))
}

#[instrument]
pub async fn greet_world() -> Html<String> {
    Html(
        html! {
            h1 { (greet_world_text().await) }
        }
        .into_string(),
    )
}

#[instrument]
pub async fn greet_world_text() -> &'static str {
    "Hello, World!"
}
